const gRDFDataSource = "rdf:localstore";
const gRDFRoot = "http://click2tab.mozdev.org/rdf/all";
const gRDFNodeClosedWindows = gRDFRoot + "/window";
const gRDFLabelClosedTabs = gRDFNodeClosedWindows + "#closedtabs";
const gRDFNodeClosedTabs = gRDFNodeClosedWindows + "/closedtabs";
const gRDFNodeSession = gRDFNodeClosedTabs + "/session";
const gRDFLabelSessionIndex = gRDFNodeSession + "#index";
const gRDFLabelSessionHistory = gRDFNodeSession + "#history";
const gRDFLabelHistoryEntry = gRDFNodeSession + "/history/entry";
const gRDFLabelHistoryEntryTitle = gRDFLabelHistoryEntry + "#title";
const gRDFLabelHistoryEntryURL = gRDFLabelHistoryEntry + "#url";
const gRDFLabelHistoryEntryScrollX = gRDFLabelHistoryEntry + "#scrollx";
const gRDFLabelHistoryEntryScrollY = gRDFLabelHistoryEntry + "#scrolly";

var ClosedTabsClick2Tab =
{
    closedTabs : [],
    cmdUndoLastClosedTab : null,
    popupsRelated : [],

    setBrowserUIState : function ()
    {
        for (var i = 0; i < this.popupsRelated.length; i++)
            TabbrowserClick2Tab.buildClosedTabList(this.popupsRelated[i]);

        if (this.closedTabs.length)
            this.cmdUndoLastClosedTab.removeAttribute("disabled");
        else
            this.cmdUndoLastClosedTab.setAttribute("disabled", "true");
    },

    cacheClosedTab : function (sessionHistory)
    {
        this.closedTabs.unshift([sessionHistory]);
        if (this.closedTabs.length > gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedCacheSize))
            this.closedTabs.pop();

        setTimeout(function () { ClosedTabsClick2Tab.setBrowserUIState(); }, 0);
    },

    removeClosedTab : function (index)
    {
        this.closedTabs.splice(index, 1);

        setTimeout(function () { ClosedTabsClick2Tab.setBrowserUIState(); }, 0);
    },

    clearAllClosedTabs : function ()
    {
        this.closedTabs.splice(0, this.closedTabs.length);

        setTimeout(function () { ClosedTabsClick2Tab.setBrowserUIState(); }, 0);
    },

    saveClosedTabs : function ()
    {
        if (!this.closedTabs.length)
            return;

        var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                                   .getService(Components.interfaces.nsIRDFService);
        var dataSource = rdfService.GetDataSource(gRDFDataSource);
        var rdfNodeClosedWindows = rdfService.GetResource(gRDFNodeClosedWindows);
        var windowContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                        .createInstance(Components.interfaces.nsIRDFContainer);
        var tabContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                     .createInstance(Components.interfaces.nsIRDFContainer);
        var containerUtils = Components.classes["@mozilla.org/rdf/container-utils;1"]
                                       .getService(Components.interfaces.nsIRDFContainerUtils);

        try
        {
            windowContainer.Init(dataSource, rdfNodeClosedWindows);
        }
        catch (ex)
        {
            windowContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedWindows);
        }

        var windowID = windowContainer.GetCount();
        var rdfNodeThisWindow = rdfService.GetResource(gRDFNodeClosedWindows + windowID);
        // Add the node to the Sequence as early as possible.
        // Not that the first element is 1 and not 0
        windowContainer.InsertElementAt(rdfNodeThisWindow, 1, true);

        var rdfLabelClosedTabs = gRDFNodeClosedWindows + windowID + "/closedtabs";
        var rdfNodeClosedTabs = rdfService.GetResource(rdfLabelClosedTabs);

        try
        {
            tabContainer.Init(dataSource, rdfNodeClosedTabs);
        }
        catch (ex)
        {
            tabContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedTabs);
        }

        for (var i = 0; i < this.closedTabs.length; i++)
        {
            var sessionHistory = this.closedTabs[i][0];

            var rdfLabelSession = rdfLabelClosedTabs + "/session" + i;
            var rdfNodeSession = rdfService.GetResource(rdfLabelSession);
            dataSource.Assert(rdfNodeSession,
                              rdfService.GetResource(gRDFLabelSessionIndex),
                              rdfService.GetIntLiteral(sessionHistory.index),
                              true);

            var rdfLabelSessionHistory = rdfLabelSession + "/history";
            var rdfNodeSessionHistory = rdfService.GetResource(rdfLabelSessionHistory);
            var entryContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                           .createInstance(Components.interfaces.nsIRDFContainer);
            try
            {
                entryContainer.Init(dataSource, rdfNodeSessionHistory);
            }
            catch (ex)
            {
                entryContainer = containerUtils.MakeSeq(dataSource, rdfNodeSessionHistory);
            }

            for (var j = 0; j < sessionHistory.count; j++)
            {
                var rdfNodeHistoryEntry = rdfService.GetResource(rdfLabelSessionHistory + "/entry" + j);
                var historyEntry = sessionHistory.getEntryAtIndex(j, false);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryTitle),
                                  rdfService.GetLiteral(historyEntry.title),
                                  true);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryURL),
                                  rdfService.GetLiteral(historyEntry.URI.spec),
                                  true);
                var scrollX = {}, scrollY = {};
                historyEntry.QueryInterface(Components.interfaces.nsISHEntry).getScrollPosition(scrollX, scrollY);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryScrollX),
                                  rdfService.GetIntLiteral(scrollX.value),
                                  true);
                dataSource.Assert(rdfNodeHistoryEntry,
                                  rdfService.GetResource(gRDFLabelHistoryEntryScrollY),
                                  rdfService.GetIntLiteral(scrollY.value),
                                  true);                
                entryContainer.AppendElement(rdfNodeHistoryEntry);
            } // end of "for (var j = 0; j < sessionHistory.count; j++)"

            dataSource.Assert(rdfNodeSession,
                              rdfService.GetResource(gRDFLabelSessionHistory),
                              rdfNodeSessionHistory,
                              true);

            tabContainer.AppendElement(rdfNodeSession);
        } // end of "for (var i = 0; i < this.closedTabs.length; i++)"
        dataSource.Assert(rdfNodeThisWindow,
                          rdfService.GetResource(gRDFLabelClosedTabs),
                          rdfNodeClosedTabs,
                          true);

        // Since it's possible that there are multiple windows open at the same time,
        // it would be better to save the RDF file immediately
        dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();
    }, // end of "saveClosedTabs : function ()"

    loadClosedTabs : function ()
    {
        var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
                                   .getService(Components.interfaces.nsIRDFService);
        var dataSource = rdfService.GetDataSource(gRDFDataSource);
        var rdfNodeClosedWindows = rdfService.GetResource(gRDFNodeClosedWindows);
        var windowContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                        .createInstance(Components.interfaces.nsIRDFContainer);
        var tabContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                     .createInstance(Components.interfaces.nsIRDFContainer);
        var containerUtils = Components.classes["@mozilla.org/rdf/container-utils;1"]
                                       .getService(Components.interfaces.nsIRDFContainerUtils);
        var ioService = Components.classes["@mozilla.org/network/io-service;1"]
                                  .getService(Components.interfaces.nsIIOService);

        try
        {
            windowContainer.Init(dataSource, rdfNodeClosedWindows);
        }
        catch (ex)
        {
            windowContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedWindows);
        }

        var windowEnum = windowContainer.GetElements();
        if (windowEnum.hasMoreElements())
        {
            var rdfNodeLastWindow = windowEnum.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
            var rdfLabelClosedTabs = rdfService.GetResource(gRDFLabelClosedTabs);
            var rdfNodeClosedTabs = dataSource.GetTarget(rdfNodeLastWindow, rdfLabelClosedTabs, true)
                                              .QueryInterface(Components.interfaces.nsIRDFResource);

            try
            {
                tabContainer.Init(dataSource, rdfNodeClosedTabs);
            }
            catch (ex)
            {
                tabContainer = containerUtils.MakeSeq(dataSource, rdfNodeClosedTabs);
            }

            var sessionEnum = tabContainer.GetElements();
            while (sessionEnum.hasMoreElements())
            {
                var sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"]
                                               .createInstance(Components.interfaces.nsISHistory);
                var sHistoryInternal = sessionHistory;
                sHistoryInternal = sHistoryInternal.QueryInterface(Components.interfaces.nsISHistoryInternal);

                var rdfNodeSession = sessionEnum.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
                var rdfLabelSessionIndex = rdfService.GetResource(gRDFLabelSessionIndex);
                var rdfNodeSessionIndex = dataSource.GetTarget(rdfNodeSession, rdfLabelSessionIndex, true)
                                                    .QueryInterface(Components.interfaces.nsIRDFInt);

                var rdfLabelSessionHistory = rdfService.GetResource(gRDFLabelSessionHistory);
                var rdfNodeSessionHistory = dataSource.GetTarget(rdfNodeSession, rdfLabelSessionHistory, true)
                                                      .QueryInterface(Components.interfaces.nsIRDFResource);
                var entryContainer = Components.classes["@mozilla.org/rdf/container;1"]
                                               .createInstance(Components.interfaces.nsIRDFContainer);

                try
                {
                    entryContainer.Init(dataSource, rdfNodeSessionHistory);
                }
                catch (ex)
                {
                    entryContainer = containerUtils.MakeSeq(dataSource, rdfNodeSessionHistory);
                }

                var entryEnum = entryContainer.GetElements();
                while (entryEnum.hasMoreElements())
                {
                    var rdfNodeHistoryEntry = entryEnum.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
                    var rdfLabelHistoryEntryTitle = rdfService.GetResource(gRDFLabelHistoryEntryTitle);
                    var rdfNodeHistoryEntryTitle = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryTitle, true)
                                                             .QueryInterface(Components.interfaces.nsIRDFLiteral);

                    var rdfLabelHistoryEntryURL = rdfService.GetResource(gRDFLabelHistoryEntryURL);
                    var rdfNodeHistoryEntryURL = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryURL, true)
                                                           .QueryInterface(Components.interfaces.nsIRDFLiteral);

                    var rdfLabelHistoryEntryScrollX = rdfService.GetResource(gRDFLabelHistoryEntryScrollX);
                    var rdfNodeHistoryEntryScrollX = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryScrollX, true)
                                                               .QueryInterface(Components.interfaces.nsIRDFInt);

                    var rdfLabelHistoryEntryScrollY = rdfService.GetResource(gRDFLabelHistoryEntryScrollY);
                    var rdfNodeHistoryEntryScrollY = dataSource.GetTarget(rdfNodeHistoryEntry, rdfLabelHistoryEntryScrollY, true)
                                                               .QueryInterface(Components.interfaces.nsIRDFInt);

                    var historyEntry = Components.classes["@mozilla.org/browser/session-history-entry;1"]
                                                 .createInstance(Components.interfaces.nsISHEntry);
                    historyEntry.SetTitle(rdfNodeHistoryEntryTitle.Value);
                    historyEntry.SetURI(ioService.newURI(rdfNodeHistoryEntryURL.Value, null, null));
                    historyEntry.setScrollPosition(rdfNodeHistoryEntryScrollX.Value, rdfNodeHistoryEntryScrollY.Value);
                    sHistoryInternal.addEntry(historyEntry, true);

                    dataSource.Unassert(rdfNodeHistoryEntry, rdfLabelHistoryEntryTitle, rdfNodeHistoryEntryTitle);
                    dataSource.Unassert(rdfNodeHistoryEntry, rdfLabelHistoryEntryURL, rdfNodeHistoryEntryURL);
                    dataSource.Unassert(rdfNodeHistoryEntry, rdfLabelHistoryEntryScrollX, rdfNodeHistoryEntryScrollX);
                    dataSource.Unassert(rdfNodeHistoryEntry, rdfLabelHistoryEntryScrollY, rdfNodeHistoryEntryScrollY);
                    entryContainer.RemoveElement(rdfNodeHistoryEntry, true);
                    entryEnum = entryContainer.GetElements();
                } // end of "while (entryEnum.hasMoreElements())"
                // delete the sequence
                this.DeleteSeq(dataSource, rdfNodeSessionHistory);

                sessionHistory.getEntryAtIndex(rdfNodeSessionIndex.Value, true);
                this.closedTabs.push([sessionHistory]);

                dataSource.Unassert(rdfNodeSession, rdfLabelSessionHistory, rdfNodeSessionHistory);
                dataSource.Unassert(rdfNodeSession, rdfLabelSessionIndex, rdfNodeSessionIndex);
                tabContainer.RemoveElement(rdfNodeSession, true);
                sessionEnum = tabContainer.GetElements();
            } // end of "while (sessionEnum.hasMoreElements())"
            // delete the sequence
            this.DeleteSeq(dataSource, rdfNodeClosedTabs);

            dataSource.Unassert(rdfNodeLastWindow, rdfLabelClosedTabs, rdfNodeClosedTabs);
            windowContainer.RemoveElement(rdfNodeLastWindow, true);

            // delete the sequence if blank
            if (!windowContainer.GetCount())
                this.DeleteSeq(dataSource, rdfNodeClosedWindows);

            dataSource.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource).Flush();

            this.setBrowserUIState();
        } // end of "if (windowEnum.hasMoreElements())"
    }, // end of loadClosedTabs : function ()"

    DeleteSeq : function (dataSource, rdfNodeSeq)
    {
        var arcOut = dataSource.ArcLabelsOut(rdfNodeSeq);
        while (arcOut.hasMoreElements())
        {
            var aLabel = arcOut.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
            var aTarget = dataSource.GetTarget(rdfNodeSeq, aLabel, true);
            dataSource.Unassert(rdfNodeSeq, aLabel, aTarget);
        }
    }
};

const Click2TabSettingsListener = 
{
    observe : function (subject, topic, prefName)
    {
        if (topic != "nsPref:changed")
            return;

        switch (prefName)
        {
            case prefStringClick2TabCloseTabButtonContext:
                TabbrowserClick2Tab.onCloseTabButtonContextChanged();
                break;
            case prefStringClick2TabReopenTabContext:
                TabbrowserClick2Tab.onReopenTabContextChanged();
                break;
            case prefStringClick2TabReopenTabsTabContext:
                TabbrowserClick2Tab.onReopenTabsTabContextChanged();
                break;
            case prefStringClick2TabCloseTabButtonMiddleClick:
                TabbrowserClick2Tab.onCloseTabButtonMiddleClickChanged();
                break;
            default:
                break;
        }
    }
};

var TabbrowserClick2Tab =
{
    prefBranch : null,

    closeTabButton : null,

    closeTabButtonContext : null,
    closeTabButtonContextReplaced : false,
    closeTabButtonOnClick : null,
    closeTabButtonOnClickReplaced : false,

    menupopupTabContex : null,
    tabContextMenuId : null,

    reopenMenuSep : null,
    reopenLastTabMenuItem : null,
    reopenTabsMenu : null,

    stripTooltipOnPopupShowing : null,

    // Since it's possbile for the UNDO menu-button to be invisible, we can't regisiter its
    // popup on initialization. But we can register the popup on the first call to its
    // "onpopupshowing" event handler. This boolean value is used to indicate whether the
    // popup has been registered to avoid multiple registration of it.
    undoButtonPopupRegistered : false,

    // Get the id of the tab bar and tab button context menu.
    // If the menu has no id, set one for it
    getTabContextMenuId : function ()
    {
        if (this.menupopupTabContex.hasAttribute("id"))
            this.tabContextMenuId = this.menupopupTabContex.getAttribute("id");
        else
        {
            this.tabContextMenuId = "tab-context-menu-click2tab";
            this.menupopupTabContex.setAttribute("id", this.tabContextMenuId);
        }
    },

    getCloseTabButtonContext : function ()
    {
        this.getTabContextMenuId();
        if (this.closeTabButton.hasAttribute("context"))
            this.closeTabButtonContext = this.closeTabButton.getAttribute("context");
        else
            this.closeTabButtonContext = this.tabContextMenuId;
    },

    replaceCloseTabButtonContext : function ()
    {
        this.closeTabButton.setAttribute("context", "undo-closetabs-popup");
        this.closeTabButtonContextReplaced = true;
    },

    restoreCloseTabButtonContext : function ()
    {
        if (this.closeTabButtonContextReplaced)
        {
            this.closeTabButton.setAttribute("context", this.closeTabButtonContext);
            this.closeTabButtonContextReplaced = false;
        }
    },

    getcloseTabButtonOnClick : function ()
    {
        if (this.closeTabButton.hasAttribute("onclick"))
            this.closeTabButtonOnClick = this.closeTabButton.getAttribute("onclick");
        else
            this.closeTabButtonOnClick = "return;";
    },

    replaceCloseTabButtonOnClick : function ()
    {
        this.closeTabButton.setAttribute("onclick", "TabbrowserClick2Tab.onCloseTabButtonClick(event);");
        this.closeTabButtonOnClickReplaced = true;
    },

    restoreCloseTabButtonOnClick : function()
    {
        if (this.closeTabButtonOnClickReplaced)
        {
            this.closeTabButton.setAttribute("onclick", this.closeTabButtonOnClick);
            this.closeTabButtonOnClickReplaced = false;
        }
    },

    addReopenToTabContext : function ()
    {
        this.reopenMenuSep = document.createElement("menuseparator");
        this.reopenMenuSep.setAttribute("hidden", "true");

        this.reopenLastTabMenuItem = document.getElementById("reopen-lastclosedtab-menuitem");
        this.reopenTabsMenu = document.getElementById("undo-closedtabs-menu");

        this.menupopupTabContex.appendChild(this.reopenMenuSep);
        this.menupopupTabContex.appendChild(this.reopenLastTabMenuItem);
        this.menupopupTabContex.appendChild(this.reopenTabsMenu);
    },

    showReopenInTabContext : function ()
    {
        this.reopenMenuSep.removeAttribute("hidden");
        this.reopenLastTabMenuItem.removeAttribute("hidden");
    },

    hideReopenInTabContext : function ()
    {
        if (this.reopenTabsMenu.hasAttribute("hidden")
            && (this.reopenTabsMenu.getAttribute("hidden") == "true"))
            this.reopenMenuSep.setAttribute("hidden", "true");
        this.reopenLastTabMenuItem.setAttribute("hidden", "true");
    },

    showReopenTabsTabContext : function ()
    {
        this.reopenMenuSep.removeAttribute("hidden");
        this.reopenTabsMenu.removeAttribute("hidden");
    },

    hideReopenTabsTabContext : function ()
    {
        if (this.reopenLastTabMenuItem.hasAttribute("hidden")
            && (this.reopenLastTabMenuItem.getAttribute("hidden") == "true"))
            this.reopenMenuSep.setAttribute("hidden", "true");
        this.reopenTabsMenu.setAttribute("hidden", "true");
    },

    removeReopenToTabContext : function ()
    {
        this.menupopupTabContex.removeChild(this.reopenMenuSep);
        this.menupopupTabContex.removeChild(this.reopenLastTabMenuItem);
        this.menupopupTabContex.removeChild(this.reopenTabsMenu);
    },

    setShortcutKeyUndo : function ()
    {
        var keyUndoLastClosedTab = document.getElementById("key_undoLastClosedTab");

        var keys = gBrowser.mPrefs.getCharPref(prefStringClick2TabShortcutKeyUndo).split("][");
        if (keys[0])
            keyUndoLastClosedTab.setAttribute("modifiers", keys[0]);
        if (keys[1])
            keyUndoLastClosedTab.setAttribute("key", keys[1]);
        if (keys[2])
            keyUndoLastClosedTab.setAttribute("keycode", keys[2]);
        if (keys[1] || keys[2])
            keyUndoLastClosedTab.setAttribute("command", "cmd_undoLastClosedTab");
    },

    removeShortcutKeyUndo : function ()
    {
        var keyUndoLastClosedTab = document.getElementById("key_undoLastClosedTab");

        if (keyUndoLastClosedTab.hasAttribute("modifiers"))
            keyUndoLastClosedTab.removeAttribute("modifiers");
        if (keyUndoLastClosedTab.hasAttribute("key"))
            keyUndoLastClosedTab.removeAttribute("key");
        if (keyUndoLastClosedTab.hasAttribute("keycode"))
            keyUndoLastClosedTab.removeAttribute("keycode");
        if (keyUndoLastClosedTab.hasAttribute("command"))
            keyUndoLastClosedTab.removeAttribute("command");
    },

    changeStripTooltip : function ()
    {
        var stripTooltip = gBrowser.mStrip.childNodes[0];

        this.stripTooltipOnPopupShowing = stripTooltip.getAttribute("onpopupshowing");
        var onStripTooltipPopupShowing = this.stripTooltipOnPopupShowing.replace(
                                            "if (document.tooltipNode.hasAttribute('label'))",
                                            "if (document.tooltipNode.hasAttribute('label') && !document.tooltipNode.hasAttribute('tooltiptext'))");
        stripTooltip.setAttribute("onpopupshowing", onStripTooltipPopupShowing);
    },

    restoreStripTooltop : function ()
    {
        var stripTooltip = gBrowser.mStrip.childNodes[0];

        stripTooltip.setAttribute("onpopupshowing", this.stripTooltipOnPopupShowing);
    },

    init : function ()
    {
        this.prefBranch = gBrowser.mPrefs.QueryInterface(Components.interfaces.nsIPrefBranchInternal);
        this.closeTabButton = document.getAnonymousElementByAttribute(gBrowser.mTabContainer,
                                                                    "class",
                                                                    "tabs-closebutton close-button");
        this.menupopupTabContex = gBrowser.mStrip.childNodes[1];

        this.getCloseTabButtonContext();
        this.getcloseTabButtonOnClick();

        this.prefBranch.addObserver(prefStringClick2TabReopenTabContext, Click2TabSettingsListener, false);
        this.prefBranch.addObserver(prefStringClick2TabReopenTabsTabContext, Click2TabSettingsListener, false);
        this.addReopenToTabContext();
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabContext))
            this.showReopenInTabContext();
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabsTabContext))
            this.showReopenTabsTabContext();

        // Change the context menu of Close Tab Button. Thanks to the idea of Twanno!
        this.prefBranch.addObserver(prefStringClick2TabCloseTabButtonContext, Click2TabSettingsListener, false);
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonContext))
            this.replaceCloseTabButtonContext();

        this.prefBranch.addObserver(prefStringClick2TabCloseTabButtonMiddleClick, Click2TabSettingsListener, false);
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonMiddleClick))
            this.replaceCloseTabButtonOnClick();

        this.setShortcutKeyUndo();
        this.changeStripTooltip();

        ClosedTabsClick2Tab.cmdUndoLastClosedTab = document.getElementById("cmd_undoLastClosedTab");
        ClosedTabsClick2Tab.popupsRelated.push(document.getElementById("undo-closetabs-popup"));
        ClosedTabsClick2Tab.popupsRelated.push(document.getElementById("undo-closetabs-menupopup"));

        // The loading of closed tabs should also be executed as late as possible for the UI
        // element to get ready
        ClosedTabsClick2Tab.loadClosedTabs();

        // It's more secure to replace those default functions in the last step of initializtion
        gBrowser.__removeTab = gBrowser.removeTab;
        gBrowser.removeTab = this.removeTab;

        __BrowserCloseTabOrWindow = BrowserCloseTabOrWindow;
        BrowserCloseTabOrWindow = this.BrowserCloseTabOrWindow;
        __WindowIsClosing = WindowIsClosing;
        WindowIsClosing = this.WindowIsClosing;
        __tryToClose = tryToClose;
        tryToClose = this.WindowIsClosing;
    },

    deinit : function ()
    {
        gBrowser.removeTab = gBrowser.__removeTab;
        BrowserCloseTabOrWindow = __BrowserCloseTabOrWindow;
        WindowIsClosing = __WindowIsClosing;

        ClosedTabsClick2Tab.saveClosedTabs();

        this.restoreStripTooltop();
        this.removeShortcutKeyUndo();
        this.restoreCloseTabButtonOnClick();
        this.restoreCloseTabButtonContext();
        this.removeReopenToTabContext();
    },

    onCloseTabButtonContextChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonContext))
            this.replaceCloseTabButtonContext();
        else
            this.restoreCloseTabButtonContext ();
    },

    onReopenTabContextChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabContext))
            this.showReopenInTabContext();
        else
            this.hideReopenInTabContext();
    },

    onReopenTabsTabContextChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenTabsTabContext))
            this.showReopenTabsTabContext();
        else
            this.hideReopenTabsTabContext();
    },

    onCloseTabButtonMiddleClickChanged : function ()
    {
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabCloseTabButtonMiddleClick))
            this.replaceCloseTabButtonOnClick();
        else
            this.restoreCloseTabButtonOnClick();
    },

    onCloseTabButtonClick : function (event)
    {
        if (event.button == 1)
            this.undoLastClosedTab();
    },

    blankTab : function (browser)
    {
        if (browser.sessionHistory.count == 0)
            return true;
        if (browser.sessionHistory.count == 1 && (browser.currentURI.spec == "about:blank"))
        {
            browser.sessionHistory.PurgeHistory(browser.sessionHistory.count);
            return true;
        }
        return false;
    },

    removeTab : function (aTab)
    {
        if (aTab.localName != "tab")
            aTab = this.mCurrentTab;

        var l = this.mTabContainer.childNodes.length;
        if (l == 1) {
            if (!this.mPrefs.getBoolPref("browser.tabs.autoHide")
                || gBrowser.mPrefs.getBoolPref(prefStringClick2TabCtlWCloseTab)
                || true/* For prefs of persistence of closed tabs*/) {
                // Clear the session history of the last tab, added by NetOne on 2004/11/16
                if (this.mPrefs.getBoolPref(prefStringClick2TabClearHistory)
                    && !TabbrowserClick2Tab.blankTab(this.mCurrentBrowser))
                {
                    // cache the session history for undo closed tabs, added by NetOne on 2004/12/12
                    ClosedTabsClick2Tab.cacheClosedTab(this.sessionHistory);
                    // webNavigation must be used since sessionHistory is a readonly property of gBrowser
                    this.webNavigation.sessionHistory = Components.classes["@mozilla.org/browser/shistory;1"]                        
                                                                  .createInstance(Components.interfaces.nsISHistory);
                }

                // blank the tab
                this.loadURI("about:blank");
                return;
            }
            // hide the tab bar
            this.mPrefs.setBoolPref("browser.tabs.forceHide", true);
            this.setStripVisibilityTo(false);
            return;
        }

        var ds = this.getBrowserForTab(aTab).docShell;

        if (ds.contentViewer && !ds.contentViewer.permitUnload())
            return;

        if (l == 2) {
            var autohide = this.mPrefs.getBoolPref("browser.tabs.autoHide");
            if (autohide)
                this.setStripVisibilityTo(false);
        }

        var index = -1;
        if (this.mCurrentTab == aTab)
            index = this.mPanelContainer.selectedIndex;
        else {
            // Find and locate the tab in our list.
            for (var i = 0; i < l; i++)
                if (this.mTabContainer.childNodes[i] == aTab)
            index = i;
        }

        // Remove the tab's filter and progress listener.
        const filter = this.mTabFilters[index];
        var oldBrowser = this.getBrowserAtIndex(index);
        oldBrowser.webProgress.removeProgressListener(filter);
        filter.removeProgressListener(this.mTabListeners[index]);
        this.mTabFilters.splice(index, 1);
        this.mTabListeners.splice(index, 1);

        // Remove our title change and blocking listeners
        oldBrowser.removeEventListener("DOMTitleChanged", this.onTitleChanged, false);

        // We are no longer the primary content area.
        oldBrowser.setAttribute("type", "content");

        // Now select the new tab before nuking the old one.
        var currentIndex = this.mPanelContainer.selectedIndex;

        var newIndex = -1;
        if (currentIndex > index)
            newIndex = currentIndex-1;
        else if (currentIndex < index)
            newIndex = currentIndex;
        else if (index == l - 1)
            newIndex = index-1;
        else
            newIndex = index;

        // cache the session history for undo closed tabs, added by NetOne on 2004/12/12
        if (!TabbrowserClick2Tab.blankTab(oldBrowser))
            ClosedTabsClick2Tab.cacheClosedTab(oldBrowser.sessionHistory);

        var oldTab = aTab;

        // clean up the before/afterselected attributes before removing the tab
        oldTab.selected = false;

        // XXX browser's destructor isn't always called, so we force a cleanup ourselves
        oldBrowser.destroy();

        this.mTabContainer.removeChild(oldTab);
        this.mPanelContainer.removeChild(this.mPanelContainer.childNodes[index]);

        this.selectedTab = this.mTabContainer.childNodes[newIndex];
        this.mPanelContainer.selectedIndex = newIndex;

        this.updateCurrentBrowser();
    },

    undoClosedTab : function (closedTabsIndex)
    {
        if (!ClosedTabsClick2Tab.closedTabs.length)
            return;

        var oldHistory = ClosedTabsClick2Tab.closedTabs[closedTabsIndex][0];
        ClosedTabsClick2Tab.removeClosedTab(closedTabsIndex);

        var aTab, aBrowser;
        if (gBrowser.webProgress.isLoadingDocument || !this.blankTab(gBrowser.mCurrentBrowser))
        {
            aTab = gBrowser.addTab("about:blank");
            if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabReopenInBackground))
                aBrowser = gBrowser.getBrowserForTab(aTab);
            else
            {
                gBrowser.selectedTab = aTab;
                aBrowser = gBrowser.mCurrentBrowser;
            }
        }
        else
            aBrowser = gBrowser.mCurrentBrowser;

        var newHistory = aBrowser.sessionHistory;
        newHistory.QueryInterface(Components.interfaces.nsISHistoryInternal);
        for (var i = 0; i < oldHistory.count; i++)
        {
            newHistory.addEntry(oldHistory.getEntryAtIndex(i, false), true);
        }
        if (oldHistory.count)
            aBrowser.webNavigation.gotoIndex(oldHistory.index);
    },

    undoLastClosedTab : function ()
    {
        this.undoClosedTab(0);
    },

    destroyClosedTabList : function (popup)
    {
        while (popup.childNodes.length > 3)
            popup.removeChild(popup.firstChild);

        var menuMorePopup = popup.childNodes[2].firstChild;
        while (menuMorePopup.childNodes.length > 2)
            menuMorePopup.removeChild(menuMorePopup.firstChild);
    },

    // This event handler doesn't need to return any value, since when it's called,
    // command "cmd_undoLastClosedTab" must have been enabled and there must be
    // at least one closed tab that needs to be displayed.
    onUndoButtonPopupShowing : function (undoButtonPopup)
    {
        if (!this.undoButtonPopupRegistered)
        {
            this.buildClosedTabList(undoButtonPopup);
            ClosedTabsClick2Tab.popupsRelated.push(undoButtonPopup);
            this.undoButtonPopupRegistered = true;
        }
    },

    buildClosedTabList : function (popup)
    {
        var menuItem;
        var i;

        this.destroyClosedTabList(popup);

        var menuSeparator = popup.childNodes[0],
            menuItemClearCache = popup.childNodes[1],
            menuMore = popup.childNodes[2];
        for (i = 0; i < gBrowser.mPrefs.getIntPref(prefStringClick2TabClosedListLength)
                        && i < ClosedTabsClick2Tab.closedTabs.length; i++)
            popup.insertBefore(this.constructTabMenuItem(i), menuSeparator);

        if (i < ClosedTabsClick2Tab.closedTabs.length)
        {
            menuItemClearCache.setAttribute("hidden", "true");
            menuMore.removeAttribute("hidden");

            var menuMorePopup = menuMore.firstChild;
            var menuMoreSeparator = menuMorePopup.childNodes[0];
            for (; i < ClosedTabsClick2Tab.closedTabs.length; i++)
                menuMorePopup.insertBefore(this.constructTabMenuItem(i), menuMoreSeparator);
        }
        else
        {
            menuItemClearCache.removeAttribute("hidden");
            menuMore.setAttribute("hidden", "true");
        }
    },

    constructTabMenuItem : function (i)
    {
        var menuItem = document.createElement("menuitem");
        var tabHistory = ClosedTabsClick2Tab.closedTabs[i][0];
        var title = tabHistory.getEntryAtIndex(tabHistory.index, false).title;
        var uri = tabHistory.getEntryAtIndex(tabHistory.index, false).URI.spec;

        var labelPrefix = "";
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabIndexInTabList))
            labelPrefix = i + ". ";
        menuItem.setAttribute("label", labelPrefix + (title || uri));
        menuItem.setAttribute("tooltiptext", title ? uri : "");
        if (gBrowser.mPrefs.getBoolPref(prefStringClick2TabIndexInTabList) && (i < 10))
            menuItem.setAttribute("accesskey", i);

        menuItem.closedTabsIndex = i;

        return menuItem;
    },

    BrowserCloseTabOrWindow : function ()
    {
        // Code changed by Netone to allow Ctrl+W only close the tab, on 12/11/2004
        if (gBrowser.localName == 'tabbrowser'
            && ((gBrowser.tabContainer.childNodes.length > 1)
                 || gBrowser.mPrefs.getBoolPref(prefStringClick2TabCtlWCloseTab))) {
            // Just close up a tab.
            gBrowser.removeCurrentTab();
            return;
        }

        // Code added by NetOne to allow tabs to be saved before the window is closing, on 01/09/2005.
        if (gBrowser.localName == 'tabbrowser')
            gBrowser.removeTab(gBrowser.mCurrentTab);
        BrowserCloseWindow();
    },

    WindowIsClosing : function()
    {
        var browser = getBrowser();
        var cn = browser.tabContainer.childNodes;
        var numtabs = cn.length;
        var reallyClose = browser.warnAboutClosingTabs(true);

        for (var i = 0; reallyClose && i < numtabs; ++i) {
            var ds = browser.getBrowserForTab(cn[i]).docShell;

            if (ds.contentViewer && !ds.contentViewer.permitUnload())
                reallyClose = false;
        }

        if (reallyClose)
        {
            // Code added by NetOne to allow tabs to be saved before the window is closing, on 01/09/2005.
            // Remove tab in reverse order so that if we undo those tabs one by one, they will be in
            // the same order as they are before the last exit.
            for (var i = numtabs - 1; i >= 0; i--)
                browser.removeTab(cn[i]);

            return closeWindow(false);
        }

        return reallyClose;
    }
};
